const ALL_SOUND_OFF_EVENT_DATA = Array.from({ length: 16 }, (_, index) => new Uint8Array([176 + index, 120, 0]));
export class MidiPlayer {
    constructor({ encodeMidiMessage, filterMidiMessage, json, midiFileSlicer, midiOutput, startIntervalScheduler, startTimeoutScheduler }) {
        this._encodeMidiMessage = encodeMidiMessage;
        this._filterMidiMessage = filterMidiMessage;
        this._json = json;
        this._midiFileSlicer = midiFileSlicer;
        this._midiOutput = midiOutput;
        this._startIntervalScheduler = startIntervalScheduler;
        this._startTimeoutScheduler = startTimeoutScheduler;
        this._state = null;
    }
    get position() {
        return this._state === null
            ? 0
            : this._state.peekScheduler === null
                ? this._state.offset
                : this._state.peekScheduler() - this._state.offset;
    }
    get state() {
        return this._state === null ? 'stopped' : this._state.peekScheduler === null ? 'paused' : 'playing';
    }
    pause() {
        if (this._state === null || this._state.peekScheduler === null) {
            throw new Error('The player is not playing.');
        }
        this._clear();
        const { endOfTrackEventTimes, resolve, peekScheduler, stopScheduler } = this._state;
        const positionWithOffset = peekScheduler();
        this._state = {
            ...this._state,
            endOfTrackEventTimes: endOfTrackEventTimes.filter((time) => time < positionWithOffset),
            offset: positionWithOffset - this._state.offset,
            peekScheduler: null,
            stopScheduler: null
        };
        stopScheduler();
        resolve();
    }
    play() {
        if (this._state !== null) {
            throw new Error('The player is not stopped.');
        }
        return this._schedule([], 0);
    }
    resume() {
        if (this._state === null || this._state.peekScheduler !== null) {
            throw new Error('The player is not paused.');
        }
        const { endOfTrackEventTimes, offset } = this._state;
        this._state = null;
        return this._schedule(endOfTrackEventTimes, offset);
    }
    stop() {
        if (this._state === null) {
            throw new Error('The player is already stopped.');
        }
        if (this._state.stopScheduler === null) {
            this._state = null;
        }
        else {
            this._clear();
            this._stop(this._state);
        }
    }
    _clear() {
        var _a, _b;
        // Bug #1: Chrome does not yet implement the clear() method.
        (_b = (_a = this._midiOutput).clear) === null || _b === void 0 ? void 0 : _b.call(_a);
        ALL_SOUND_OFF_EVENT_DATA.forEach((data) => this._midiOutput.send(data));
    }
    _schedule(endOfTrackEventTimes, offset) {
        return new Promise((resolve) => {
            var _a;
            const [peekScheduler, stopScheduler] = this._startIntervalScheduler(({ end, start }) => {
                var _a, _b, _c, _d, _e;
                if (this._state === null) {
                    this._state = { endOfTrackEventTimes, offset: start - offset, resolve, peekScheduler: null, stopScheduler: null };
                }
                const events = this._midiFileSlicer.slice(start - this._state.offset, end - this._state.offset);
                events
                    .filter(({ event }) => this._filterMidiMessage(event))
                    .forEach(({ event, time }) => this._midiOutput.send(this._encodeMidiMessage(event), start + time));
                const newEndOfTrackEventTimes = events
                    .filter(({ event }) => MidiPlayer._isEndOfTrack(event))
                    .map(({ time }) => start + time);
                this._state.endOfTrackEventTimes.push(...newEndOfTrackEventTimes);
                if (this._state.endOfTrackEventTimes.length === this._json.tracks.length) {
                    const timeout = Math.max(...newEndOfTrackEventTimes) - ((_c = (_b = (_a = this._state).peekScheduler) === null || _b === void 0 ? void 0 : _b.call(_a)) !== null && _c !== void 0 ? _c : start);
                    if (timeout > 0) {
                        (_e = (_d = this._state).stopScheduler) === null || _e === void 0 ? void 0 : _e.call(_d);
                        this._state = {
                            ...this._state,
                            stopScheduler: this._startTimeoutScheduler(() => {
                                this._state = null;
                                resolve();
                            }, timeout)
                        };
                    }
                    else {
                        this._stop(this._state);
                    }
                }
            });
            if (((_a = this._state) === null || _a === void 0 ? void 0 : _a.stopScheduler) === null) {
                this._state = { ...this._state, peekScheduler, stopScheduler };
            }
            else {
                stopScheduler();
                if (this._state !== null) {
                    this._state = { ...this._state, peekScheduler };
                }
            }
        });
    }
    _stop({ resolve, stopScheduler }) {
        this._state = null;
        stopScheduler === null || stopScheduler === void 0 ? void 0 : stopScheduler();
        resolve();
    }
    static _isEndOfTrack(event) {
        return 'endOfTrack' in event;
    }
}
//# sourceMappingURL=midi-player.js.map